import io
import uuid

import pytest
from pydantic import SecretStr
from testcontainers_on_whales.minio import MinioContainer

from aio_microservice import Service, ServiceSettings, http, startup_hook
from aio_microservice.http import TestHttpClient
from aio_microservice.s3 import S3Extension, S3ExtensionSettings, S3Settings


async def test_s3_client() -> None:
    class TestSettings(ServiceSettings, S3ExtensionSettings):
        bucket_name: str = "test-bucket"

    class TestService(Service[TestSettings], S3Extension):
        @startup_hook
        async def create_s3_bucket(self) -> None:
            self.s3.client.create_bucket(Bucket=self.settings.bucket_name)

        @http.post(path="/test-upload")
        async def post_upload(self, content: str) -> str:
            file_id = uuid.uuid4()
            file_obj = io.BytesIO(content.encode())
            self.s3.client.upload_fileobj(
                Fileobj=file_obj,
                Bucket=self.settings.bucket_name,
                Key=f"{file_id}.txt",
            )
            return str(file_id)

    with MinioContainer() as container:
        container.wait_ready(timeout=120)

        settings = TestSettings(
            s3=S3Settings(
                endpoint_url=container.get_connection_url(),
                access_key_id=container.username,
                secret_access_key=SecretStr(container.password),
            ),
        )

        service = TestService(settings=settings)

        async with TestHttpClient(service=service) as http_client:
            response_post = await http_client.post("/test-upload?content=TEST-CONTENT")
            assert response_post.status_code == http.status_codes.HTTP_201_CREATED

            response_list_objects = service.s3.client.list_objects(
                Bucket=service.settings.bucket_name,
            )
            object_keys = [c["Key"] for c in response_list_objects["Contents"]]
            assert f"{response_post.text}.txt" in object_keys


async def test_s3_connection_verification_on_startup(
    caplog: pytest.LogCaptureFixture,
) -> None:
    class TestSettings(ServiceSettings, S3ExtensionSettings): ...

    class TestService(Service[TestSettings], S3Extension): ...

    settings = TestSettings(
        s3=S3Settings(
            endpoint_url="http://localhost:12345",
            access_key_id="somerandomid",
            secret_access_key=SecretStr("somerandomsecret"),
        ),
    )

    service = TestService(settings=settings)

    async with TestHttpClient(service=service):
        pass

    assert "Failed to verify connection" in caplog.text


async def test_s3_readiness_probe() -> None:
    class TestSettings(ServiceSettings, S3ExtensionSettings): ...

    class TestService(Service[TestSettings], S3Extension): ...

    container = MinioContainer()
    container.start()
    container.wait_ready(timeout=120)

    settings = TestSettings(
        s3=S3Settings(
            endpoint_url=container.get_connection_url(),
            access_key_id=container.username,
            secret_access_key=SecretStr(container.password),
        ),
    )

    service = TestService(settings=settings)

    async with TestHttpClient(service=service) as http_client:
        response_running = await http_client.get("/readiness")

        container.stop()
        container.wait_exited(timeout=10)

        response_stopped = await http_client.get("/readiness")

        assert response_running.status_code == http.status_codes.HTTP_200_OK
        assert response_stopped.status_code == http.status_codes.HTTP_503_SERVICE_UNAVAILABLE



from aio_microservice import (
    Service,
    ServiceSettings,
)


async def test_custom_settings() -> None:
    class TestSettings(ServiceSettings):
        test_value: str = "TEST"

    class TestService(Service[TestSettings]): ...

    service = TestService()
    assert isinstance(service.settings, TestSettings)
    assert service.settings.test_value == "TEST"



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, ClassVar

import pytest
from testcontainers_on_whales.rabbitmq import RabbitmqContainer

from aio_microservice import Service, ServiceSettings, amqp, http
from aio_microservice.amqp import (
    AmqpExtension,
    AmqpExtensionSettings,
    AmqpSettings,
    BaseMiddleware,
    TestAmqpBroker,
)
from aio_microservice.http import TestHttpClient

if TYPE_CHECKING:
    from collections.abc import Iterable

    from pytest_mock import MockerFixture


@pytest.fixture
def rabbitmq_container() -> Iterable[RabbitmqContainer]:
    with RabbitmqContainer() as container:
        container.wait_ready(timeout=120)
        yield container


@pytest.fixture
def rabbitmq_ip(rabbitmq_container: RabbitmqContainer) -> str:
    return rabbitmq_container.get_container_ip()


@pytest.fixture
def rabbitmq_port(rabbitmq_container: RabbitmqContainer) -> int:
    return rabbitmq_container.get_container_port(RabbitmqContainer.RABBITMQ_PORT)


async def test_amqp_readiness_probe() -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension): ...

    container = RabbitmqContainer()
    container.start()
    container.wait_ready(timeout=120)

    settings = TestSettings(
        amqp=AmqpSettings(
            host=container.get_container_ip(),
            port=container.get_container_port(RabbitmqContainer.RABBITMQ_PORT),
        ),
    )
    service = TestService(settings=settings)

    async with TestHttpClient(service=service) as http_client:
        response_running = await http_client.get("/readiness")

        container.stop()
        container.wait_exited(timeout=10)

        response_stopped = await http_client.get("/readiness")

        assert response_running.status_code == http.status_codes.HTTP_200_OK
        assert response_stopped.status_code == http.status_codes.HTTP_503_SERVICE_UNAVAILABLE


async def test_amqp_retry_on_failure(
    mocker: MockerFixture,
    rabbitmq_ip: str,
    rabbitmq_port: int,
) -> None:
    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        def __init__(self, settings: TestSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.handler_stub = mocker.stub()
            self.handler_should_fail = True

        @amqp.subscriber(queue="test-subscriber-queue", retry=True)
        async def handle_test(self, message: str) -> None:
            self.handler_stub(message)
            if self.handler_should_fail:
                self.handler_should_fail = False
                raise Exception()  # noqa: TRY002

    settings = TestSettings(amqp=AmqpSettings(host=rabbitmq_ip, port=rabbitmq_port))
    service = TestService(settings=settings)

    async with TestAmqpBroker(service=service, with_real=True) as amqp_broker:
        await amqp_broker.publish(queue="test-subscriber-queue", message="TEST")
        await asyncio.sleep(0.5)

        assert len(service.handler_stub.call_args_list) == 2


async def test_amqp_middleware(mocker: MockerFixture) -> None:
    middleware_stub = mocker.stub()

    class TestMiddleware(BaseMiddleware):
        async def on_receive(self) -> None:
            middleware_stub()
            await super().on_receive()

    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        __amqp_middlewares__: ClassVar = [TestMiddleware]

        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self, message: str) -> None:
            pass

    service = TestService()

    async with TestAmqpBroker(service) as amqp_broker:
        await amqp_broker.publish(queue="test-subscriber-queue", message="TEST")

    middleware_stub.assert_called_once()


async def test_amqp_prefetch_count_none(mocker: MockerFixture) -> None:
    handler_pre_stub = mocker.stub()
    handler_post_stub = mocker.stub()

    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        def __init__(self, settings: TestSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.keep_running = True

        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self) -> None:
            handler_pre_stub()
            while self.keep_running:
                await asyncio.sleep(0.1)
            handler_post_stub()

    service = TestService()

    async with TestAmqpBroker(service) as amqp_broker:
        task1 = asyncio.create_task(amqp_broker.publish(queue="test-subscriber-queue"))
        task2 = asyncio.create_task(amqp_broker.publish(queue="test-subscriber-queue"))

        await asyncio.sleep(0.5)

        assert handler_pre_stub.call_count == 2
        assert handler_post_stub.call_count == 0

        service.keep_running = False

        await task1
        await task2

        assert handler_pre_stub.call_count == 2
        assert handler_post_stub.call_count == 2


async def test_amqp_prefetch_count_one(
    mocker: MockerFixture,
    rabbitmq_ip: str,
    rabbitmq_port: int,
) -> None:
    handler_pre_stub = mocker.stub()
    handler_post_stub = mocker.stub()

    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        __amqp_prefetch_count__ = 1

        def __init__(self, settings: TestSettings | None = None) -> None:
            super().__init__(settings=settings)
            self.keep_running = True

        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self) -> None:
            handler_pre_stub()
            while self.keep_running:
                await asyncio.sleep(0.1)
            handler_post_stub()

    settings = TestSettings(amqp=AmqpSettings(host=rabbitmq_ip, port=rabbitmq_port))
    service = TestService(settings=settings)

    async with TestAmqpBroker(service, with_real=True) as amqp_broker:
        task1 = asyncio.create_task(amqp_broker.publish(queue="test-subscriber-queue"))
        task2 = asyncio.create_task(amqp_broker.publish(queue="test-subscriber-queue"))

        await asyncio.sleep(0.5)

        assert handler_pre_stub.call_count == 1
        assert handler_post_stub.call_count == 0

        service.keep_running = False

        await asyncio.sleep(0.5)

        await task1
        await task2

        assert handler_pre_stub.call_count == 2
        assert handler_post_stub.call_count == 2


async def test_amqp_graceful_timeout_none(
    mocker: MockerFixture,
    rabbitmq_ip: str,
    rabbitmq_port: int,
) -> None:
    handler_pre_stub = mocker.stub()
    handler_post_stub = mocker.stub()
    handler_cancelled_stub = mocker.stub()

    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self) -> None:
            handler_pre_stub()
            try:
                await asyncio.sleep(0.5)
            except asyncio.CancelledError:
                handler_cancelled_stub()
            else:
                handler_post_stub()

    settings = TestSettings(amqp=AmqpSettings(host=rabbitmq_ip, port=rabbitmq_port))
    service = TestService(settings=settings)

    publish_task = None

    async with TestAmqpBroker(service=service, with_real=True) as amqp_broker:
        publish_task = asyncio.create_task(amqp_broker.publish(queue="test-subscriber-queue"))

        await asyncio.sleep(0.1)

        assert handler_pre_stub.call_count == 1
        assert handler_post_stub.call_count == 0

    await publish_task

    assert handler_cancelled_stub.call_count == 1
    assert handler_pre_stub.call_count == 1
    assert handler_post_stub.call_count == 0


async def test_amqp_graceful_timeout_lower_than_handler_duration(
    mocker: MockerFixture,
    rabbitmq_ip: str,
    rabbitmq_port: int,
) -> None:
    handler_pre_stub = mocker.stub()
    handler_post_stub = mocker.stub()
    handler_cancelled_stub = mocker.stub()

    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        __amqp_graceful_timeout__ = 0.1

        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self) -> None:
            handler_pre_stub()
            try:
                await asyncio.sleep(0.5)
            except asyncio.CancelledError:
                handler_cancelled_stub()
            else:
                handler_post_stub()

    settings = TestSettings(amqp=AmqpSettings(host=rabbitmq_ip, port=rabbitmq_port))
    service = TestService(settings=settings)

    publish_task = None

    async with TestAmqpBroker(service=service, with_real=True) as amqp_broker:
        publish_task = asyncio.create_task(amqp_broker.publish(queue="test-subscriber-queue"))

        await asyncio.sleep(0.1)

        assert handler_pre_stub.call_count == 1
        assert handler_post_stub.call_count == 0

    await publish_task

    assert handler_cancelled_stub.call_count == 1
    assert handler_pre_stub.call_count == 1
    assert handler_post_stub.call_count == 0


async def test_amqp_graceful_timeout_higher_handler_duration(
    mocker: MockerFixture,
    rabbitmq_ip: str,
    rabbitmq_port: int,
) -> None:
    handler_pre_stub = mocker.stub()
    handler_post_stub = mocker.stub()
    handler_cancelled_stub = mocker.stub()

    class TestSettings(ServiceSettings, AmqpExtensionSettings): ...

    class TestService(Service[TestSettings], AmqpExtension):
        __amqp_graceful_timeout__ = 1.0

        @amqp.subscriber(queue="test-subscriber-queue")
        async def handle_test(self) -> None:
            handler_pre_stub()
            try:
                await asyncio.sleep(0.5)
            except asyncio.CancelledError:
                handler_cancelled_stub()
            else:
                handler_post_stub()

    settings = TestSettings(amqp=AmqpSettings(host=rabbitmq_ip, port=rabbitmq_port))
    service = TestService(settings=settings)

    publish_task = None

    async with TestAmqpBroker(service=service, with_real=True) as amqp_broker:
        publish_task = asyncio.create_task(amqp_broker.publish(queue="test-subscriber-queue"))

        await asyncio.sleep(0.1)

        assert handler_pre_stub.call_count == 1
        assert handler_post_stub.call_count == 0

    await publish_task

    assert handler_cancelled_stub.call_count == 0
    assert handler_pre_stub.call_count == 1
    assert handler_post_stub.call_count == 1



from litestar import Controller, delete, get, patch, post, put, status_codes

from aio_microservice.core.service import HttpSettings
from aio_microservice.core.testing import TestHttpClient

__all__ = [
    "Controller",
    "HttpSettings",
    "TestHttpClient",
    "delete",
    "get",
    "patch",
    "post",
    "put",
    "status_codes",
]



from aio_microservice import Service, ServiceSettings, http
from aio_microservice.http import TestHttpClient


async def test_http_openapi_swagger_ui() -> None:
    class TestService(Service[ServiceSettings]):
        @http.get(path="/test")
        async def get_test(self) -> str:
            return "TEST"

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/schema/openapi")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert "/test" in response.text


async def test_http_openapi_json() -> None:
    class TestService(Service[ServiceSettings]):
        @http.get(path="/test")
        async def get_test(self) -> str:
            return "TEST"

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/schema/openapi.json")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert "/test" in response.json()["paths"]


async def test_http_openapi_yaml() -> None:
    class TestService(Service[ServiceSettings]):
        @http.get(path="/test")
        async def get_test(self) -> str:
            return "TEST"

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/schema/openapi.yaml")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert "/test" in response.text


async def test_http_openapi_service_description() -> None:
    class TestService(Service[ServiceSettings]):
        __description__ = "TEST SERVICE DESCRIPTION"

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/schema/openapi.json")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert response.json()["info"]["description"] == TestService.__description__


async def test_http_openapi_service_version() -> None:
    class TestService(Service[ServiceSettings]):
        __version__ = "1.2.3"

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/schema/openapi.json")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert response.json()["info"]["version"] == TestService.__version__



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Callable, ClassVar, TypeVar

from faststream import BaseMiddleware, FastStream
from faststream.rabbit import RabbitBroker, RabbitExchange, RabbitQueue
from loguru import logger
from pydantic import BaseModel, Field, SecretStr
from typing_extensions import Concatenate, ParamSpec

from aio_microservice.amqp.asyncapi import make_asyncapi_controller
from aio_microservice.core.abc import (
    ExtensionABC,
    readiness_probe,
    shutdown_hook,
    startup_hook,
    startup_message,
)
from aio_microservice.types import Port  # noqa: TCH001

if TYPE_CHECKING:
    from faststream.rabbit.shared.schemas import ReplyConfig
    from faststream.rabbit.shared.types import TimeoutType


class AmqpSettings(BaseModel):
    host: str = Field(
        default="localhost",
        description="The ip for connecting to the broker.",
    )
    port: Port = Field(
        default=5672,
        description="The port for connecting to the broker.",
    )
    username: str = Field(
        default="guest",
        description="The username for authentication on the broker.",
    )
    password: SecretStr = Field(
        default=SecretStr("guest"),
        description="The password for authentication on the broker.",
    )


class AmqpExtensionImpl:
    def __init__(self, service: AmqpExtension, settings: AmqpSettings) -> None:
        self._service = service
        self._settings = settings
        self._faststream_rabbit_broker = RabbitBroker(
            host=self._settings.host,
            port=self._settings.port,
            login=self._settings.username,
            password=self._settings.password.get_secret_value(),
            middlewares=service.__amqp_middlewares__,
            max_consumers=service.__amqp_prefetch_count__,
            graceful_timeout=service.__amqp_graceful_timeout__,
        )
        self._faststream_app = FastStream(
            broker=self._faststream_rabbit_broker,
            title=service.__class__.__name__,
            version=service.__version__,
            description=service.__description__,
        )
        self._asyncapi_controller = make_asyncapi_controller(
            app=self._faststream_app,
        )
        self._register_handlers()

    def _register_handlers(self) -> None:
        handler_methods = inspect.getmembers(
            object=self._service,
            predicate=lambda obj: hasattr(obj, AmqpDecorator.MARKER),
        )
        for handler_name, _ in handler_methods:
            handler = getattr(self._service, handler_name)
            handler_settings = getattr(handler, AmqpDecorator.MARKER)
            for handler_setting in handler_settings:
                if isinstance(handler_setting, subscriber):
                    wrapper = self._faststream_rabbit_broker.subscriber(
                        queue=handler_setting.queue,
                        exchange=handler_setting.exchange,
                        reply_config=handler_setting.reply_config,
                        no_ack=handler_setting.no_ack,
                        retry=handler_setting.retry,
                        title=handler_setting.title,
                        description=handler_setting.description,
                        include_in_schema=handler_setting.include_in_schema,
                    )
                elif isinstance(handler_setting, publisher):  # pragma: no branch
                    wrapper = self._faststream_rabbit_broker.publisher(
                        queue=handler_setting.queue,
                        exchange=handler_setting.exchange,
                        routing_key=handler_setting.routing_key,
                        reply_to=handler_setting.reply_to,
                        mandatory=handler_setting.mandatory,
                        immediate=handler_setting.immediate,
                        persist=handler_setting.persist,
                        timeout=handler_setting.timeout,
                        priority=handler_setting.priority,
                        title=handler_setting.title,
                        description=handler_setting.description,
                        include_in_schema=handler_setting.include_in_schema,
                    )
                handler = wrapper(handler)

    @property
    def broker(self) -> RabbitBroker:
        return self._faststream_rabbit_broker


class AmqpExtensionSettings(BaseModel):
    amqp: AmqpSettings = AmqpSettings()


class AmqpExtension(ExtensionABC):
    __amqp_middlewares__: ClassVar[list[type[BaseMiddleware]]] = []
    __amqp_prefetch_count__: ClassVar[int | None] = None
    __amqp_graceful_timeout__: ClassVar[float | None] = None

    def __init__(self, settings: AmqpExtensionSettings) -> None:
        self.amqp = AmqpExtensionImpl(service=self, settings=settings.amqp)
        self.register_http_controller(self.amqp._asyncapi_controller)

    @startup_message
    async def _amqp_startup_message(self) -> str:
        return """
            # Amqp

            The following http-endpoints are available:
            - {scheme}://{host}:{port}/schema/asyncapi
            - {scheme}://{host}:{port}/schema/asyncapi.json
            - {scheme}://{host}:{port}/schema/asyncapi.yaml
        """

    @startup_hook
    async def _amqp_startup_hook(self) -> None:
        logger.info("Connecting to broker")
        await self.amqp._faststream_rabbit_broker.start()

    @shutdown_hook
    async def _amqp_shutdown_hook(self) -> None:
        logger.info("Disconnecting from broker")
        await self.amqp._faststream_rabbit_broker.close()

    @readiness_probe
    async def _amqp_readiness_probe(self) -> bool:
        if self.amqp._faststream_rabbit_broker._connection:
            # NOTE this might break.
            # faststream maintainer says, they will provide functionality like this.
            return self.amqp._faststream_rabbit_broker._connection.connected.is_set()
        return False  # pragma: no cover


AmqpExtensionT = TypeVar("AmqpExtensionT", bound=AmqpExtension)
P = ParamSpec("P")
R = TypeVar("R")


class AmqpDecorator:
    MARKER = "_amqp_decorator"


# TODO support all faststream decorator options
class subscriber(AmqpDecorator):  # noqa: N801
    def __init__(
        self,
        queue: RabbitQueue | str,
        exchange: RabbitExchange | str | None = None,
        reply_config: ReplyConfig | None = None,
        no_ack: bool = False,
        retry: bool | int = False,
        # AsyncAPI information
        title: str | None = None,
        description: str | None = None,
        include_in_schema: bool = True,
    ) -> None:
        self.queue = queue
        self.exchange = exchange
        self.reply_config = reply_config
        self.no_ack = no_ack
        self.retry = retry
        self.title = title
        self.description = description
        self.include_in_schema = include_in_schema

    def __call__(
        self,
        fn: Callable[Concatenate[AmqpExtensionT, P], R],
    ) -> Callable[Concatenate[AmqpExtensionT, P], R]:
        handlers = getattr(fn, self.MARKER, [])
        handlers.append(self)
        setattr(fn, self.MARKER, handlers)
        return fn


class publisher(AmqpDecorator):  # noqa: N801
    def __init__(
        self,
        queue: RabbitQueue | str = "",
        exchange: RabbitExchange | str | None = None,
        routing_key: str = "",
        reply_to: str | None = None,
        mandatory: bool = True,
        immediate: bool = False,
        persist: bool = False,
        timeout: TimeoutType = None,
        priority: int | None = None,
        # AsyncAPI information
        title: str | None = None,
        description: str | None = None,
        include_in_schema: bool = True,
    ) -> None:
        self.queue = queue
        self.exchange = exchange
        self.routing_key = routing_key
        self.reply_to = reply_to
        self.mandatory = mandatory
        self.immediate = immediate
        self.persist = persist
        self.timeout = timeout
        self.priority = priority
        self.title = title
        self.description = description
        self.include_in_schema = include_in_schema

    def __call__(
        self,
        fn: Callable[Concatenate[AmqpExtensionT, P], R],
    ) -> Callable[Concatenate[AmqpExtensionT, P], R]:
        handlers = getattr(fn, self.MARKER, [])
        handlers.append(self)
        setattr(fn, self.MARKER, handlers)
        return fn



from aio_microservice import Service, ServiceSettings, http
from aio_microservice.http import TestHttpClient


async def test_http_test_client() -> None:
    class TestService(Service[ServiceSettings]):
        @http.get(path="/test")
        async def get_test(self) -> str:
            return "TEST"

    service = TestService()
    async with TestHttpClient(service=service) as client:
        response = await client.get("/test")
        assert response.status_code == http.status_codes.HTTP_200_OK
        assert response.text == "TEST"



import multiprocessing
import re

import httpx
import pytest
from pydantic import Field
from pytest_mock import MockFixture

from aio_microservice import (
    Service,
    ServiceSettings,
    http,
)


def test_cli_help(capsys: pytest.CaptureFixture[str], mocker: MockFixture) -> None:
    class TestService(Service[ServiceSettings]): ...

    mocker.patch.dict("os.environ", {"NO_COLOR": "1", "TERM": "dumb"})
    mocker.patch("sys.argv", ["test-service", "--help"])

    with pytest.raises(SystemExit):
        TestService.cli()

    captured = capsys.readouterr()
    assert "Usage: test-service [OPTIONS]" in captured.out


def test_cli_help_custom_settings(
    capsys: pytest.CaptureFixture[str],
    mocker: MockFixture,
) -> None:
    class TestSettings(ServiceSettings):
        test_value: str = Field(default="TEST", description="TEST DESCRIPTION")

    class TestService(Service[TestSettings]): ...

    mocker.patch.dict("os.environ", {"NO_COLOR": "1", "TERM": "dumb"})
    mocker.patch("sys.argv", ["test-service", "--help"])

    with pytest.raises(SystemExit):
        TestService.cli()

    captured = capsys.readouterr()
    assert "Usage: test-service [OPTIONS]" in captured.out
    matches = re.findall(r"--test-value\s+TEXT\s+TEST DESCRIPTION", captured.out)
    assert len(matches) == 1, "custom setting not found in help text"


def test_cli_help_service_description(
    capsys: pytest.CaptureFixture[str],
    mocker: MockFixture,
) -> None:
    class TestService(Service[ServiceSettings]):
        __description__ = "TEST SERVICE DESCRIPTION"

    mocker.patch.dict("os.environ", {"NO_COLOR": "1", "TERM": "dumb"})
    mocker.patch("sys.argv", ["test-service", "--help"])

    with pytest.raises(SystemExit):
        TestService.cli()

    captured = capsys.readouterr()
    assert "Usage: test-service [OPTIONS]" in captured.out
    assert "TEST SERVICE DESCRIPTION" in captured.out


def test_cli_run(mocker: MockFixture) -> None:
    class TestService(Service[ServiceSettings]): ...

    mocker.patch.dict("os.environ", {"NO_COLOR": "1", "TERM": "dumb"})
    mocker.patch("sys.argv", ["test-service", "--http-port=1234"])
    p = multiprocessing.Process(target=TestService.cli)
    p.start()

    transport = httpx.HTTPTransport(retries=5)
    client = httpx.Client(transport=transport)

    response = client.get("http://localhost:1234/readiness")
    assert response.status_code == http.status_codes.HTTP_200_OK

    p.terminate()
    p.join()



from aio_microservice.core.abc import (
    lifespan_hook,
    liveness_probe,
    readiness_probe,
    shutdown_hook,
    startup_hook,
    startup_message,
)
from aio_microservice.core.service import (
    Service,
    ServiceSettings,
)

__all__ = [
    "Service",
    "ServiceSettings",
    "lifespan_hook",
    "liveness_probe",
    "readiness_probe",
    "shutdown_hook",
    "startup_hook",
    "startup_message",
]



from __future__ import annotations

import asyncio
import logging
import typing
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from functools import partial
from textwrap import dedent
from typing import Any, Callable, ClassVar, Generic, TypeVar

import litestar
import rich
import rich_click as click
import typed_settings
import uvicorn
from humps import kebabize
from loguru import logger
from pydantic import BaseModel, Field

from aio_microservice.core.abc import ServiceABC, startup_message
from aio_microservice.core.logging import setup_logging
from aio_microservice.core.openapi import OpenAPIController
from aio_microservice.types import Port  # noqa: TCH001

if typing.TYPE_CHECKING:
    from collections.abc import AsyncGenerator


class HttpSettings(BaseModel):
    host: str = Field(
        default="localhost",
        description="The ip to bind to.",
    )
    port: Port = Field(
        default=8080,
        description="The port to bind to.",
    )


class ServiceSettings(BaseModel):
    debug: bool = Field(
        default=False,
        description="Whether to enable debug logging.",
    )
    http: HttpSettings = HttpSettings()


ServiceSettingsT = TypeVar("ServiceSettingsT", bound=ServiceSettings)


class Service(Generic[ServiceSettingsT], ServiceABC):
    _settings_cls: ClassVar[type[ServiceSettings]]

    def __init_subclass__(cls) -> None:
        # determine settings-class based on generic
        cls._settings_cls = typing.get_args(cls.__orig_bases__[0])[0]  # type: ignore

    def __init__(self, settings: ServiceSettingsT | None = None) -> None:
        if settings is None:
            settings = typing.cast(ServiceSettingsT, self._settings_cls())
        ServiceABC.__init__(self, settings=settings)

        self.settings = settings

        self._litestar_app = self._create_litestar_app()
        self._uvicorn_server = self._create_uvicorn_server(self._litestar_app)

    @property
    def litestar_app(self) -> litestar.Litestar:
        return self._litestar_app

    def _get_litestar_on_startup(self) -> list[litestar.types.LifespanHook]:
        fns: list[litestar.types.LifespanHook] = [partial(fn, self) for fn in self._startup_hooks]
        fns.append(self._emit_startup_message)
        return fns

    def _get_litestar_on_shutdown(self) -> list[litestar.types.LifespanHook]:
        return [partial(fn, self) for fn in self._shutdown_hooks]

    def _get_litestar_lifespan(
        self,
    ) -> list[[Callable[litestar.Litestar], AbstractAsyncContextManager[Any]]]:  # type: ignore
        # NOTE this is ugly, but works
        @asynccontextmanager
        async def wrapper(fn: Any, app: litestar.Litestar) -> AsyncGenerator[None, None]:  # noqa: ANN401
            fn_ctx = asynccontextmanager(fn)
            async with fn_ctx(self):
                yield

        return [partial(wrapper, hook.fn) for hook in self._lifespan_hooks]

    def _get_litestar_route_handlers(
        self,
    ) -> list[litestar.types.ControllerRouterHandler]:
        litestar_route_handlers = self._litestar_http_controllers
        for route_handler in self._litestar_http_route_handlers:
            # replace fn with a partial to emulate Controller behavior
            route_handler._fn = partial(route_handler.fn, self)
            litestar_route_handlers.append(route_handler)

        def create_probe_response(value: bool) -> litestar.Response[bool]:
            status_code = (
                litestar.status_codes.HTTP_200_OK
                if value
                else litestar.status_codes.HTTP_503_SERVICE_UNAVAILABLE
            )
            return litestar.Response(content=value, status_code=status_code)

        @litestar.get(path="/readiness", include_in_schema=False)
        async def _get_readiness() -> litestar.Response[bool]:
            readiness = await self._get_readiness()
            return create_probe_response(readiness)

        @litestar.get(path="/liveness", include_in_schema=False)
        async def _get_liveness() -> litestar.Response[bool]:
            readiness = await self._get_liveness()
            return create_probe_response(readiness)

        litestar_route_handlers.append(_get_readiness)
        litestar_route_handlers.append(_get_liveness)
        return litestar_route_handlers

    def _create_litestar_app(self) -> litestar.Litestar:
        openapi_config = litestar.openapi.OpenAPIConfig(
            title=self.__class__.__name__,
            version=self.__version__,
            description=self.__description__,
            enabled_endpoints={"openapi.json", "openapi.yaml", "openapi.yml"},
            openapi_controller=OpenAPIController,
        )
        return litestar.Litestar(
            on_startup=self._get_litestar_on_startup(),
            on_shutdown=self._get_litestar_on_shutdown(),
            lifespan=self._get_litestar_lifespan(),
            route_handlers=self._get_litestar_route_handlers(),
            listeners=self._litestar_listeners,
            openapi_config=openapi_config,
        )

    def _create_uvicorn_server(self, litestar_app: litestar.Litestar) -> uvicorn.Server:
        config = uvicorn.Config(
            app=litestar_app,
            host=self.settings.http.host,
            port=self.settings.http.port,
            log_config=None,
            log_level=logging.DEBUG if self.settings.debug else logging.INFO,
        )
        return uvicorn.Server(
            config=config,
        )

    async def _get_readiness(self) -> bool:
        fns = [fn(self) for fn in self._readiness_probes]
        results = await asyncio.gather(*fns)
        return all(results)

    async def _get_liveness(self) -> bool:
        fns = [fn(self) for fn in self._liveness_probes]
        results = await asyncio.gather(*fns)
        return all(results)

    async def _emit_startup_message(self) -> None:
        messages = [await fn(service=self) for fn in self._startup_messages]
        if not messages:
            return  # pragma: no cover
        startup_message = ""
        for index, message in enumerate(messages):
            _message = dedent(message).lstrip().rstrip()
            startup_message += _message.format(
                scheme="https" if self._uvicorn_server.config.is_ssl else "http",
                host=self._uvicorn_server.config.host,
                port=self._uvicorn_server.config.port,
            )
            if index < len(messages) - 1:
                startup_message += "\n\n"

        console = rich.console.Console()
        markdown = rich.markdown.Markdown(startup_message)
        console.print(markdown)

    @startup_message
    async def _startup_message(self) -> str:
        return """
            # Http

            The following http-endpoints are available:
            - {scheme}://{host}:{port}/schema/openapi
            - {scheme}://{host}:{port}/schema/openapi.json
            - {scheme}://{host}:{port}/schema/openapi.yaml
        """

    async def run(self) -> None:
        logger.info(f"Starting Service: {self.__class__.__name__}")
        logger.info(f"Using Settings: {self.settings}")
        await self._uvicorn_server.serve()

    @classmethod
    def _cli_run(cls, settings: ServiceSettingsT) -> None:
        loglevel = logging.DEBUG if settings.debug else logging.INFO
        setup_logging(level=loglevel)
        service = cls(settings=settings)
        asyncio.run(service.run())

    @classmethod
    def cli(cls) -> None:
        click_options = typed_settings.click_options(
            settings_cls=cls._settings_cls,
            loaders=kebabize(cls.__name__),
            show_envvars_in_help=True,
        )
        wrapped = click_options(lambda settings: cls._cli_run(settings=settings))
        click_wrapper = click.command(help=cls.__description__)
        click_command = click_wrapper(wrapped)
        click_command()
